Unit Real3D;

{ WolfEnstein clone - with 3D Texture mapping

Copyright (C) 1997-98 TothPaul@mygale.org
 http://www.mygale.org/~tothpaul

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


First Release 3/97

v1 - 2D mapping
v2 - 3D mapping, doors and joystick
   - no more REAL but precalculed sin & cos.
v3 - less ASM but now support VESA mode 101h (640x480x256)
   - adding VIDEOSWP.PAS - remove XMODE :(
   - fix door drawing (back face)
v4 - adding polygonal Portal (each room draw is own floor & roof)
   - back to XMODE - VESA is realy to slow on my 486 + Oak Tech. card :)
   - fix bugs (integer:=longint can change sign !)
v5 - Delphi 2 (Windows) version 1
   - remove lot of ASM cause there's something I don't understand in Delphi BASM :(
   - back to REAL :( to avoid arithmetics overflow (previously handled by BASM)
}

Interface

{$I MYOPT.PAS}


Uses
 Windows,
 MyDef;

Const
 Size=64;
{ each wall has a picture of 64x64 pixels - this have to change in futur version }
 PictureSize=64*64;
{ 256 RGB colors }
 PaletteSize=3*256;

Type
 TDib=array[0..1024*768] of byte; // dummy size to support 1024*768 :)
 TPalette=Array[0..255,0..2] of byte;

Var
 zoom:integer; // equal to screenwidth

 Columns:word;  { for statistic only }
 Portals:word;

 RoomCount   :word;
 Rooms       :^TRooms;
 PictureCount:word;
 Pictures    :^TPictures;

 Palette     :^TPalette;
 Page        :byte;
 PageOfs     :word;

 PlayerX     :Integer;
 PlayerY     :Integer;
 PlayerR     :Integer;
 PlayerA     :byte;
 PlayerH     :integer; { altitude }

 Psin:real; { Player Sin & Cos }
 Pcos:real;
 Wsin:real;
 Wcos:real;

 GetMaxX:integer;
 LineSize:integer;
 GetMaxY:integer;
 HalfX:integer; { Image Middle = Screen less StatusBar }
 HalfY:integer;

 WPalette:HPalette;
 DIB:^TDib;

Procedure StartGame; { init Graphics... }
Procedure ShowView;  { Render the view }
Procedure StopGame;
procedure ResetSize(width,height:integer);

Implementation

Var
{ Current Wall pos }
 WallX  :real;
 WallZ  :real;
{ Parameters for DrawWall - can be global }
 Picture:PPicture;

 oa,oc:longint;  { TextureMapping vars }
 ha,hc:longint;
  a, c:longint;

  u, v:integer;  { Texture[u,v] }

 xd,hd:integer;
 ix,ih:integer;
  i, e:integer;
     h:integer;
   max:integer;

 video  :longint;
 deltay :integer;
 f1,f2  :integer;
 y      :integer;

 VideoX1:real;
 VideoX2:real;
 vx1,vx2:integer;
 VideoH1:real;
 VideoH2:real;

 WallTop1   :real;
 WallTop2   :real;
 WallBottom1:real;
 WallBottom2:real;

 RoomH  :integer; { current Room altitude }

 StepH  :integer;
 StepWay:integer;
 StepH1 :real;
 StepH2 :real;

Type
 TSlice=array[0..1024] of word; // for a maximum screen resolution of 1024x768

Var
 ClipTop   :TSlice;
 ClipBottom:TSlice;
 WallTop   :TSlice;
 WallBottom:TSlice;
 StepPos   :TSlice;

Procedure FillWord(Var Data; Count,Value:dword); assembler;
{ eax = @Data
  edx = Count
  ecx = Value }
 asm
  mov ebx,edi    // save edi (needed with Delphi)
   mov edi,eax   // get @data
   mov eax,ecx   // get value
   mov ecx,edx   // get count
   rep stosw
  mov edi,ebx
 (* dos version *
  les di,Data
  mov cx,Count
  mov ax,Value
  rep stosw
 *)
 end;

Procedure StartGame;
 var
  r:real;
  i:word;
  Pal256:pLogPalette;
 begin
{ Create Logical Palette }
 GetMem(Pal256,4+4*256);
 with Pal256^ do begin
  palVersion:=$300;
  palNumEntries:=256;
 { convert DOS RGB64 palette to Windows RGB256 palette }
  for i:=0 to 255 do
   with palPalEntry[i] do begin
    peRed  :=Palette^[i,0] shl 2;
    peGreen:=Palette^[i,1] shl 2;
    peBlue :=Palette^[i,2] shl 2;
    peFlags:=PC_NOCOLLAPSE;
   end;
 end;
{ Create Palette }
 WPalette:=CreatePalette(Pal256^);
 FreeMem(Pal256,4+4*256);

end;

Procedure Scan(Var Slices:TSlice;VideoH1,VideoH2:longint);
 begin
 { compute Wall Limite (top & bottom) for current wall }
  xd:=vx2-vx1; //round(Videox2-Videox1); { allways from left to right }
  hd:=Videoh2-Videoh1; if hd<0 then begin hd:=-hd; ih:=-1 end else ih:=1;

  if Videoh1<-16000 then Videoh1:=-16000; { fix bug :) }
  if Videoh1>+16000 then Videoh1:=+16000; { 16000 is a dummy value }

  ix:=vx1;
   h:=Videoh1+HalfY;

  if xd>=hd then begin

   f1:= 2*hd;
   e :=f1-xd;
   f2:= e-xd;
   for i:=VX1 to VX2 do begin
    Slices[ix]:=h;
    inc(ix);
    if e<0 then inc(e,f1) else begin
     inc(e,f2);
     inc(h,ih);
    end;
   end;

  end else begin

   f1:= 2*xd;
   e :=f1-hd;
   f2:= e-hd;
   for i:=0 to hd do begin
    if e<0 then inc(e,f1) else begin
     inc(e,f2);
     Slices[ix]:=h;
     inc(ix);
     if ix>VideoX2 then break;
     Slices[ix]:=h; { to be sure that next Slice will be set }
    end;
    inc(h,ih);
   end;
  end;

 end;

Procedure VLine(x,y,height,color:word);
 var
  video:longint;
  h:integer;
 begin
  video:=LineSize*(GetMaxY-y)+x;
  for h:=1 to height do begin
   dib[video]:=color;
   dec(video,LineSize); // remember that DIBs are upsidedown !
  end;
 end;

Var
 Floor,Roof:word;

Procedure RoofAndFloor;
 begin
 { roof & floor }
  for ix:=Vx1 to Vx2 do begin

   { Roof }
    y:=ClipTop[ix];
    h:=WallTop[ix];
    if h>ClipBottom[ix] then h:=ClipBottom[ix];
    if h>y then begin
     ClipTop[ix]:=h;
     dec(h,y);
     VLine(ix,y,h,Roof);
    end;

   { floor }
    y:=WallBottom[ix];
    if y<ClipTop[ix] then y:=ClipTop[ix];
    h:=ClipBottom[ix];
    if h>y then begin
     ClipBottom[ix]:=y;
     dec(h,y);
     VLine(ix,y,h,Floor);
    end;

  end;
 end;

Procedure StepUp(minx,maxx:word);
 begin
  for ix:=minx to maxx do begin
   y:=WallTop[ix];
   if y<ClipTop[ix] then y:=ClipTop[ix];
   h:=StepPos[ix];
   if h>ClipBottom[ix] then h:=ClipBottom[ix];
   if h>y then begin
    ClipTop[ix]:=h;
    dec(h,y);
    VLine(ix,y,h,Roof-1);
   end;
  end;
 end;

Procedure StepDown(minx,maxx:word);
 begin
  for ix:=minx to maxx do begin
   y:=StepPos[ix];
   if y<ClipTop[ix] then y:=ClipTop[ix];
   h:=WallBottom[ix];
   if h>ClipBottom[ix] then h:=ClipBottom[ix];
   if h>y then begin
    ClipBottom[ix]:=y;
    dec(h,y);
    VLine(ix,y,h,Floor+1);
   end;
  end;
 end;

Procedure DrawWall(minx,maxx,width:integer);
 begin
  c:=oc+hc*(minx-HalfX);
  a:=oa+ha*(minx-HalfX);
  for ix:=minx to maxx do begin

   if (c>0)and(ClipTop[ix]<ClipBottom[ix]) then begin

    inc(Columns);
    u:=(Width*a div c) and 63;

    y:=WallTop[ix];
    deltay:=WallBottom[ix]-y;
    v:=0;

   { draw wall }
    Video:=LineSize*(GetMaxY-y)+ix;

    if deltay>=63 then begin

     f1:=2*63;
     e :=f1-deltay;
     f2:=e -deltay;

     if y<ClipTop[ix] then begin
      for i:=y+1 to ClipTop[ix] do begin
       if e<0 then inc(e,f1) else begin
        inc(e,f2);
        inc(v);
       end;
      end;
      y:=ClipTop[ix];
      Video:=LineSize*(GetMaxY-y)+ix;
     end;

     max:=WallBottom[ix];
     if max>Clipbottom[ix] then max:=ClipBottom[ix];

     for i:=y to max-1 do begin
      dib[Video]:=Picture^[u,v];
      dec(video,LineSize);
      if e<0 then inc(e,f1) else begin
       inc(e,f2);
       inc(v);
      end;
     end;

    end else begin

     f1:=2*deltay;
     e :=f1-63;
     f2:=e -63;

     if y<ClipTop[ix] then begin
      max:=ClipTop[ix];
      for i:=0 to max do begin
       if e<0 then inc(e,f1) else begin
        inc(y); if y>=max then break;
        inc(e,f2);
       end;
       inc(v);
      end;
      y:=ClipTop[ix];
      Video:=LineSize*(GetMaxY-y)+ix;
     end;

     max:=ClipBottom[ix];

     for i:=v to 63 do begin
      if e<0 then inc(e,f1) else begin
       dib[Video]:=Picture^[u,i];
       inc(y); if y>=max then break;
       dec(video,LineSize);
       inc(e,f2);
      end;
     end;

    end;
   end; { c>0 }

   inc(c,hc);
   inc(a,ha);

  end; { next column (x) }

 end;


Procedure DrawPortal(Room:word;MinX,MaxX:Integer);
 Var
  Wall  :byte;

  NextX :real;
  NextZ :real;

 ClipX1 :real;
 ClipZ1 :real;
 ClipX2 :real;
 ClipZ2 :real;

  x1,x2 :longint;

  min   :longint;

 begin
  Inc(Portals); // stat
  x1:=-1;
  With (Rooms^[Room]^) do begin
   NextX:=PlayerX-Left; { room hook }
   NextZ:=PlayerY-Top;

   WallX:=(NextZ*Psin)+(-NextX*Pcos);   { room rotation }
   WallZ:=(NextZ*PCos)-(-NextX*Psin)+1;

   For Wall:=1 to WallCount do begin { follow walls }

    With Walls[Wall] do begin { for this wall, where is his other side }

     wSin:=Width*sin((PlayerA-Axis)*PI/128);
     wCos:=Width*cos((PlayerA-Axis)*PI/128);
     NextX:=WallX+wSin;
     NextZ:=WallZ+wCos;

     if (WallZ>1)or(NextZ>1) then begin { at least one side is in front }

      { Left Z Clipping }
      if (WallZ<1) then begin
       ClipX1:=NextX-(NextX-WallX)*(NextZ) / (NextZ-WallZ);
       ClipZ1:=1;
      end else begin
       ClipX1:=WallX;
       ClipZ1:=WallZ;
      end;

      { Right Z Clipping }
      if NextZ<1 then begin
       ClipX2:=WallX+(NextX-WallX)*(WallZ) / (WallZ-NextZ);
       ClipZ2:=1;
      end else begin
       ClipX2:=NextX;
       ClipZ2:=NextZ;
      end;

     { Screen X positions }
      VideoX1:=zoom*ClipX1 / ClipZ1+HalfX;
      VideoX2:=zoom*ClipX2 / ClipZ2+HalfX;

     { Range Clipping - is this wall in front of the screen }
      if (VideoX1<VideoX2)and(VideoX1<=MaxX)and(VideoX2>=MinX) then begin
       RoomH:=Height;
      { something to keep the "stair" heigth }
       if Door=0 then
        StepWay:=0    { no door = no step }
       else begin
        StepH  :=Rooms^[Door]^.Height;
        StepWay:=RoomH-StepH;   { way to draw from current room height }
        if StepWay<>0 then begin
         StepH  :=PlayerH-StepH; { where to draw from Player's room }
         if StepWay<0 then begin
          StepH1:=zoom*(StepH+32) / ClipZ1;
          StepH2:=zoom*(StepH+32) / ClipZ2;
         end else begin
          StepH1:=zoom*(StepH-32) / ClipZ1;
          StepH2:=zoom*(StepH-32) / ClipZ2;
         end;
        end;
       end;

       RoomH:=PlayerH-RoomH;

      { unlike previous version, I need Top & Left, not just Height }
       WallTop1:=zoom*(RoomH-32) / ClipZ1;
       WallBottom1:=zoom*(RoomH+32) / ClipZ1;

       WallTop2:=zoom*(RoomH-32) / ClipZ2;
       WallBottom2:=zoom*(RoomH+32) / ClipZ2;

      { Left X Clipping }
       if VideoX1<MinX then begin
        WallTop1:=WallTop2-(WallTop2-WallTop1)*(VideoX2-MinX) / (VideoX2-VideoX1);
        WallBottom1:=WallBottom2-(WallBottom2-WallBottom1)*(VideoX2-MinX) / (VideoX2-VideoX1);
        if StepWay<>0 then
         StepH1 := StepH2-( StepH2- StepH1)*(VideoX2-MinX) / (VideoX2-VideoX1);
        VideoX1:=MinX;
       end;

      { Right X Clipping }
       if VideoX2>MaxX then begin
        WallTop2:=WallTop1+(WallTop2-WallTop1)*(MaxX-VideoX1) / (VideoX2-VideoX1);
        WallBottom2:=WallBottom1+(WallBottom2-WallBottom1)*(MaxX-VideoX1) / (VideoX2-VideoX1);
        if StepWay<>0 then
         StepH2 := StepH1+( StepH2- StepH1)*(MaxX-VideoX1) / (VideoX2-VideoX1);
        VideoX2:=MaxX;
       end;

      { before recurrent call }
       x1:=round(VideoX1); vx1:=x1;
       x2:=round(VideoX2); vx2:=x2;

      { update clipping before recursive DrawPortal }
       Scan(WallTop,round(WallTop1),round(WallTop2));
       Scan(WallBottom,round(WallBottom1),round(WallBottom2));

       Roof :=16*1+Room*2;
       Floor:=16*6+Room*2;
       RoofAndFloor;

       { Check Portal }
       if (Door>0) then begin
        if StepWay<>0 then begin
         Scan(StepPos,round(StepH1),round(StepH2));
         if StepWay>0 then StepUp(x1,x2) else StepDown(x1,x2);
        end;
        DrawPortal(Door,x1,x2);
       end; { Portal }

       if (Image<>0) then begin
        Picture:=@Pictures[Image];
        oa:=round(-Size*ClipX1);
        oc:=round( Size*(ClipX2-ClipX1));
        ha:=round( Size*ClipZ1 / zoom);
        hc:=round(-Size*(ClipZ2-ClipZ1) / zoom);
        DrawWall(X1,X2,Width);
       end;

      end; { Range Clipping }

     end; { Z Clipping }

    end; { With Walls[iWall] }

    WallX:=NextX;
    WallZ:=NextZ;

   end; { Next Wall }

  end; { This Room }

 end;

Procedure ShowView;
begin
  { Rotation (in DrawPortal) }

  {
         Z                   Z
         :                   :    z               x =Left-playerx
    Wall :                   : A /\               z =playery-Top
      +--:------x            :  /  \
      |  :                   : /    \Wall         x'=(z.sin a)+(x.cos a)
      |  :                   :/     /             z'=(z.cos a)-(x.sin a)+K
      |  +.....X       Player+...../.....X
      | Player               \ A  /
      |                      :\  /
      y                      : \/
                             :  x
                             :
                             +---------- K
  }

 { statictics }
  Columns:=0;
  Portals:=0;
 { InitClipping }
  FillWord(ClipTop,GetMaxX,0);
  FillWord(ClipBottom,GetMaxX,GetMaxY+1);
 { Init Height }
  RoomH:=Rooms^[PlayerR]^.Height;
 { fall down }
  if PlayerH>RoomH then dec(PlayerH) else PlayerH:=RoomH;
 { draw rooms }
  DrawPortal(PlayerR,0,GetMaxX-1); { recurrent procedure }

end;

Procedure StopGame;
 begin
  DeleteObject(wPalette);
  FreeMem(Dib,LineSize*GetMaxY);
 end;

procedure ResetSize(width,height:integer);
 begin
  if dib<>nil then FreeMem(Dib,LineSize*GetMaxY);
  LineSize:=(width+3) and $FFFC;
  GetMaxX:=width;
  GetMaxY:=height;
  HalfX:=GetMaxX shr 1;
  HalfY:=GetMaxY shr 1;
  GetMem(Dib,LineSize*GetMaxY);
 end;

end.